// Copyright (C) 2023, Xilinx Inc - All rights reserved
//
// Licensed under the Apache License, Version 2.0 (the "License"). You may
// not use this file except in compliance with the License. A copy of the
// License is located at
//
//      http://www.apache.org/licenses/LICENSE-2.0
//
// Unless required by applicable law or agreed to in writing, software
// distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
// WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
// License for the specific language governing permissions and limitations
// under the License.

#include "jsf_mamgmt.h"

#include <sys/stat.h>

#include <string_view>

#include <Logging.h>

#include <jsf_helpers.h>

#include <Device.h>
#include <VfioDevice.h>
#include <TranscoderDevice.h>
#include <Telemetry.h>
#include <Flash.h>
#include <Sriov.h>
#include <Reset.h>

using namespace nlohmann::json_literals;
using namespace std::literals;

namespace {

std::unique_ptr<jsf_int::Jsf> s_jsf;

class DeviceWrapper : public jsf_int::NodeInternal {
private:
  std::unique_ptr<Device> m_dev;

public:
  explicit DeviceWrapper(const pci_dev& PciDevice, bool Demo = false) : NodeInternal(jsf_helpers::DbdfString(PciDevice)), m_dev(Device::CreateDevice(PciDevice, Demo)) {
    s_jsf->Insert(this, s_jsf->Create<Reset>(*m_dev, "reset")); // Reset
    if (PciDevice.device_id == 0x5071) {                        // Hypervisor: mamgmt Telemetry for VF device is Not available
      return;
    }

    s_jsf->Insert(this, s_jsf->Create<Sriov>(*m_dev, "numvfs")); // Sriov (testapps/jmamgmt --live devices/BDF/numvfs '"value":0')

    s_jsf->Insert(this, s_jsf->Create<Flash>(*m_dev, "flash")); // Flash

    // Telemetry
    auto devNode = s_jsf->Insert(this, s_jsf->Create<jsf_int::NodeInternal>("device_info"s)); // MA35 Device Info

    s_jsf->Insert(devNode, s_jsf->Create<jsf_int::NodeJson>("processor_type", "VPU (Video Processing Unit)"));
    s_jsf->Insert(devNode, s_jsf->Create<ProductInfoNode>(*m_dev, "product_name"s, 6, Device::NAME));                   // product_info -> name
    s_jsf->Insert(devNode, s_jsf->Create<ProductInfoNode>(*m_dev, "product_revision"s, 2, Device::REVISION));           // product_info -> revision
    s_jsf->Insert(devNode, s_jsf->Create<ProductInfoNode>(*m_dev, "product_serial_number"s, 4, Device::SERIAL_NUMBER)); // product_info -> serial_number
    // s_jsf->Insert(devNode, s_jsf->Create<ProductInfoNode>(*this, "max_power_mode"s, 9, 0x0));

    auto chipIdNode = s_jsf->Insert(devNode, s_jsf->Create<jsf_int::NodeInternal>("chip_id"s));                // MA35 HW Id's
    s_jsf->Insert(chipIdNode, s_jsf->Create<ProductInfoNode>(*m_dev, "part_number"s, 6, Device::PART_NUMBER)); // product_info -> part_number
    s_jsf->Insert(chipIdNode, s_jsf->Create<SkuNumber>(*m_dev, "sku_number"s, 6, Device::PART_NUMBER));        // Last two digit of part_number
    s_jsf->Insert(chipIdNode, s_jsf->Create<ProductInfoNode>(*m_dev, "oem_id"s, 1, Device::OEM_ID));           // product_info -> oem_id
    // s_jsf->Insert(chipIdNode, s_jsf->Create<jsf_int::NodeJson>("chip_type", ""));

    auto thresholdNode = s_jsf->Insert(devNode, s_jsf->Create<jsf_int::NodeInternal>("thresholds"s)); // MA35 Thresholds
    s_jsf->Insert(thresholdNode, s_jsf->Create<jsf_int::NodeJson>("threshold_temp_C", 85));           // BMC ASIC Junction
    s_jsf->Insert(thresholdNode, s_jsf->Create<jsf_int::NodeJson>("max_operating_temp_C", 105));      // Warning Limit ASIC Junction
    s_jsf->Insert(thresholdNode, s_jsf->Create<jsf_int::NodeJson>("shutdown_temp_C", 110));           // Fatal Limit ASIC Junction

    auto verNode = s_jsf->Insert(devNode, s_jsf->Create<jsf_int::NodeInternal>("version_info"s)); // MA35 Version Info
    s_jsf->Insert(verNode, s_jsf->Create<VersionInfo>(*m_dev, "ZSP"s, 0, Device::VERSION_INFO));
    s_jsf->Insert(verNode, s_jsf->Create<VersionInfo>(*m_dev, "SC"s, 1, Device::VERSION_INFO));
    s_jsf->Insert(verNode, s_jsf->Create<VersionInfo>(*m_dev, "eSecure"s, 2, Device::VERSION_INFO));
    s_jsf->Insert(verNode, s_jsf->Create<VersionInfo>(*m_dev, "PCIe"s, 3, Device::VERSION_INFO));
    s_jsf->Insert(verNode, s_jsf->Create<VersionInfo>(*m_dev, "PcieCtlPatch"s, 4, Device::VERSION_INFO));
    s_jsf->Insert(verNode, s_jsf->Create<VersionInfo>(*m_dev, "PciePhyPatch"s, 5, Device::VERSION_INFO));

    // auto pcieNode = s_jsf->Insert(this, s_jsf->Create<jsf_helpers::NodePcieInfo>("pcie"s, PciDevice)); // PCIe Info
    auto pcieNode = s_jsf->Insert(this, s_jsf->Create<jsf_int::NodeInternal>("pcie_info"s)); // Pcie Info
    s_jsf->Insert(pcieNode, s_jsf->Create<PcieInfo>(*m_dev, "vendor_id"s, "vendor"s));
    s_jsf->Insert(pcieNode, s_jsf->Create<PcieInfo>(*m_dev, "device_id"s, "device"s));
    s_jsf->Insert(pcieNode, s_jsf->Create<PcieInfo>(*m_dev, "speed"s, "current_link_speed"s));
    s_jsf->Insert(pcieNode, s_jsf->Create<PcieInfo>(*m_dev, "width"s, "current_link_width"s));
#if 0
    s_jsf->Insert(pcieNode, s_jsf->Create<VariableValNode>(*m_dev, "tx_bw_MBps"s, "pcie_rc2ep_bw"s));
    s_jsf->Insert(pcieNode, s_jsf->Create<VariableValNode>(*m_dev, "rx_bw_MBps"s, "pcie_ep2rc_bw"s));
#endif

    s_jsf->Insert(devNode, s_jsf->Create<Uptime>(*m_dev, "uptime_S"s, Device::UPTIME_S)); // Uptime

    auto telNode = s_jsf->Insert(this, s_jsf->Create<jsf_int::NodeInternal>("telemetry"s)); // Telemetry

#if 0
    auto deviceMemNode = s_jsf->Insert(telNode, s_jsf->Create<jsf_int::NodeInternal>("device_memory"s)); // Device Memory
    s_jsf->Insert(deviceMemNode, s_jsf->Create<DeviceMemory>(*m_dev, "region0"s, 0, "device_mem_info"s, Demo));
    s_jsf->Insert(deviceMemNode, s_jsf->Create<DeviceMemory>(*m_dev, "region1"s, 1, "device_mem_info"s, Demo));
#endif

    auto ddrBwNode = s_jsf->Insert(telNode, s_jsf->Create<jsf_int::NodeInternal>("ddr_bandwidth"s)); // DDR Bandwidth
    s_jsf->Insert(ddrBwNode, s_jsf->Create<DdrInfoNode>(*m_dev, "s1_axi_r_MBps"s, 0, Device::DDR_BW_INFO));
    s_jsf->Insert(ddrBwNode, s_jsf->Create<DdrInfoNode>(*m_dev, "s1_axi_w_MBps"s, 1, Device::DDR_BW_INFO));
    s_jsf->Insert(ddrBwNode, s_jsf->Create<DdrInfoNode>(*m_dev, "s1_dfi_r_MBps"s, 2, Device::DDR_BW_INFO));
    s_jsf->Insert(ddrBwNode, s_jsf->Create<DdrInfoNode>(*m_dev, "s1_dfi_w_MBps"s, 3, Device::DDR_BW_INFO));
    s_jsf->Insert(ddrBwNode, s_jsf->Create<DdrInfoNode>(*m_dev, "s2_axi_r_MBps"s, 4, Device::DDR_BW_INFO));
    s_jsf->Insert(ddrBwNode, s_jsf->Create<DdrInfoNode>(*m_dev, "s2_axi_w_MBps"s, 5, Device::DDR_BW_INFO));
    s_jsf->Insert(ddrBwNode, s_jsf->Create<DdrInfoNode>(*m_dev, "s2_dfi_r_MBps"s, 6, Device::DDR_BW_INFO));
    s_jsf->Insert(ddrBwNode, s_jsf->Create<DdrInfoNode>(*m_dev, "s2_dfi_w_MBps"s, 7, Device::DDR_BW_INFO));

    auto errCounterNode = s_jsf->Insert(telNode, s_jsf->Create<jsf_int::NodeInternal>("error_counters"));             // Error Counters
    s_jsf->Insert(errCounterNode, s_jsf->Create<ErrorCounters>(*m_dev, "pcie", 0, Device::PCIE_ERROR_COUNTER));       // PcieErrorCounters
    s_jsf->Insert(errCounterNode, s_jsf->Create<ErrorCounters>(*m_dev, "ddr_ch0", 0, Device::DDR_ECC_ERROR_COUNTER)); // DDRErrorCounters
    s_jsf->Insert(errCounterNode, s_jsf->Create<ErrorCounters>(*m_dev, "ddr_ch1", 1, Device::DDR_ECC_ERROR_COUNTER));
    s_jsf->Insert(errCounterNode, s_jsf->Create<ErrorCounters>(*m_dev, "ddr_ch2", 2, Device::DDR_ECC_ERROR_COUNTER));
    s_jsf->Insert(errCounterNode, s_jsf->Create<ErrorCounters>(*m_dev, "ddr_ch3", 3, Device::DDR_ECC_ERROR_COUNTER));
    s_jsf->Insert(errCounterNode, s_jsf->Create<ErrorCounters>(*m_dev, "ddr_ch4", 4, Device::DDR_ECC_ERROR_COUNTER));
    s_jsf->Insert(errCounterNode, s_jsf->Create<ErrorCounters>(*m_dev, "ddr_ch5", 5, Device::DDR_ECC_ERROR_COUNTER));
    s_jsf->Insert(errCounterNode, s_jsf->Create<ErrorCounters>(*m_dev, "ddr_ch6", 6, Device::DDR_ECC_ERROR_COUNTER));
    s_jsf->Insert(errCounterNode, s_jsf->Create<ErrorCounters>(*m_dev, "ddr_ch7", 7, Device::DDR_ECC_ERROR_COUNTER));
    s_jsf->Insert(errCounterNode, s_jsf->Create<ErrorCounters>(*m_dev, "THS1_axi_sram", 0, Device::AXI_SRAM_ECC_ERROR_COUNTER)); // SramErrorCounters
    s_jsf->Insert(errCounterNode, s_jsf->Create<ErrorCounters>(*m_dev, "THS2_axi_sram", 1, Device::AXI_SRAM_ECC_ERROR_COUNTER));

    // Chip level Telemetry
    auto asicNode = s_jsf->Insert(telNode, s_jsf->Create<jsf_int::NodeInternal>("chip"s)); // sn_chip

    auto voltNode = s_jsf->Insert(asicNode, s_jsf->Create<jsf_int::NodeInternal>("pvt_voltage")); // Chip Voltage
    s_jsf->Insert(voltNode, s_jsf->Create<DeviceVoltage>(*m_dev, "sensor0"s, "enc"s, "volt_mV", 0, Device::PVT_VOLTAGE, Demo));
    s_jsf->Insert(voltNode, s_jsf->Create<DeviceVoltage>(*m_dev, "sensor1"s, "ml_engine"s, "volt_mV", 1, Device::PVT_VOLTAGE, Demo));
    s_jsf->Insert(voltNode, s_jsf->Create<DeviceVoltage>(*m_dev, "sensor2"s, "ddr0"s, "volt_mV", 2, Device::PVT_VOLTAGE, Demo));
    s_jsf->Insert(voltNode, s_jsf->Create<DeviceVoltage>(*m_dev, "sensor3"s, "aux"s, "volt_mV", 3, Device::PVT_VOLTAGE, Demo));

    auto tempNode = s_jsf->Insert(asicNode, s_jsf->Create<jsf_int::NodeInternal>("pvt_temperature")); // Chip Temperature
    s_jsf->Insert(tempNode, s_jsf->Create<DeviceTemperature>(*m_dev, "sensor0"s, "ma35_temp_s0"s, "temp_C", 0, Device::PVT_TEMPERATURE, Demo));
    s_jsf->Insert(tempNode, s_jsf->Create<DeviceTemperature>(*m_dev, "sensor1"s, "ma35_temp_s1"s, "temp_C", 1, Device::PVT_TEMPERATURE, Demo));
    s_jsf->Insert(tempNode, s_jsf->Create<DeviceTemperature>(*m_dev, "sensor2"s, "ma35_temp_s2"s, "temp_C", 2, Device::PVT_TEMPERATURE, Demo));
    s_jsf->Insert(tempNode, s_jsf->Create<DeviceTemperature>(*m_dev, "sensor3"s, "ma35_temp_s3"s, "temp_C", 3, Device::PVT_TEMPERATURE, Demo));

    // Board level Telemetry
    auto boardNode = s_jsf->Insert(telNode, s_jsf->Create<jsf_int::NodeInternal>("board"s));

    auto bTemprNode = s_jsf->Insert(boardNode, s_jsf->Create<jsf_int::NodeInternal>("temperature")); // Board Temperature
    s_jsf->Insert(bTemprNode, s_jsf->Create<SensorInfoNode>(*m_dev, "board_temp"s, "board_temp"s, "temp_C", 0, Device::BOARD_TEMP, Demo));

    auto bPowerNode = s_jsf->Insert(boardNode, s_jsf->Create<jsf_int::NodeInternal>("power")); // Board Power
    s_jsf->Insert(bPowerNode, s_jsf->Create<BoardPower>(*m_dev, "board_power"s, "board_power"s, "power_mW", 0, Device::BOARD_VC, Demo));

    auto bVoltNode = s_jsf->Insert(boardNode, s_jsf->Create<jsf_int::NodeInternal>("voltage_current")); // board_vc
    s_jsf->Insert(bVoltNode, s_jsf->Create<BoardVcNode>(*m_dev, "sensor0"s, "12V PEX Current"s, "current_mA", 0, Device::BOARD_VC, Demo));
    s_jsf->Insert(bVoltNode, s_jsf->Create<BoardVcNode>(*m_dev, "sensor1"s, "3V AUX Current"s, "current_mA", 1, Device::BOARD_VC, Demo));
    s_jsf->Insert(bVoltNode, s_jsf->Create<BoardVcNode>(*m_dev, "sensor2"s, "3V PEX Current"s, "current_mA", 2, Device::BOARD_VC, Demo));
    s_jsf->Insert(bVoltNode, s_jsf->Create<BoardVcNode>(*m_dev, "sensor3"s, "12V PEX Voltage"s, "volt_mV", 3, Device::BOARD_VC, Demo));
    s_jsf->Insert(bVoltNode, s_jsf->Create<BoardVcNode>(*m_dev, "sensor4"s, "3V AUX Voltage"s, "volt_mV", 4, Device::BOARD_VC, Demo));
    s_jsf->Insert(bVoltNode, s_jsf->Create<BoardVcNode>(*m_dev, "sensor5"s, "3V PEX Voltage"s, "volt_mV", 5, Device::BOARD_VC, Demo));
  }
}; // namespace

void DemoInit(jsf_int::Node* Node) {
  s_jsf->Insert("version"sv, s_jsf->Create<jsf_int::NodeJson>("pkg_version", "0.0.1"));

  srand(time(nullptr));
  pci_dev dummy   = {};
  dummy.vendor_id = 0x10ee;
  dummy.device_id = 0x5070;

  dummy.bus = 4;
  s_jsf->Insert(Node, s_jsf->Create<DeviceWrapper>(dummy, true));
  dummy.bus = 5;
  s_jsf->Insert(Node, s_jsf->Create<DeviceWrapper>(dummy, true));

  static const std::vector<const char*> sDevices = {"0000:04:00.0", "0000:05:00.0"};
  s_jsf->Insert(""sv, s_jsf->Create<jsf_int::NodeJson>("device_list", sDevices));
}

} // namespace

namespace jsf_mamgmt {

void init(bool Demo) {
  s_jsf = jsf_int::CreateJsf();

  s_jsf->Insert(""sv, s_jsf->Create<jsf_helpers::NodeSystemInfo>("host"));
  auto devicesNode = s_jsf->Insert(""sv, s_jsf->Create<jsf_int::NodeInternal>("devices"));

  if (Demo) {
    DemoInit(devicesNode);
    return;
  }

  s_jsf->Insert("version"sv, s_jsf->Create<jsf_helpers::NodeVersionInfo>("pkg_version", "/opt/amd/ama/ma35/version"));

#ifndef VFIO
  std::vector<std::string> pciDevices;
  std::vector<std::string> deviceList;
  for (unsigned dev_cnt = 0; dev_cnt < jsf_helpers::MAX_TRANS_DEVICES; ++dev_cnt) {
    std::ifstream stream("/sys/class/misc/ama_transcoder"s + std::to_string(dev_cnt) + "/bus_id");
    if (stream.is_open()) {
      std::string transBdf;
      std::getline(stream, transBdf);
      pciDevices.push_back(transBdf);
    }
  }
#endif

  auto devices = jsf_helpers::GetPciDevices(0x10ee, 0x5070);
  for (const auto& device : devices) {
#ifndef VFIO
    auto dbdf = jsf_helpers::DbdfString(device);
    if (std::find(pciDevices.begin(), pciDevices.end(), dbdf) != pciDevices.end()) {
      s_jsf->Insert(devicesNode, s_jsf->Create<DeviceWrapper>(device));
      deviceList.push_back(dbdf);
    } else {
      logNotice("Transcoder device does not exist: %s", dbdf.c_str());
    }
#else
    s_jsf->Insert(devicesNode, s_jsf->Create<DeviceWrapper>(device));
    deviceList.push_back(dbdf);
#endif
  }

  auto devices_vf = jsf_helpers::GetPciDevices(0x10ee, 0x5071);
  for (const auto& device : devices_vf) {
    s_jsf->Insert(devicesNode, s_jsf->Create<DeviceWrapper>(device));
  }
  std::ranges::transform(devices_vf, std::back_inserter(deviceList), jsf_helpers::DbdfString);
  s_jsf->Insert(""sv, s_jsf->Create<jsf_int::NodeJson>("device_list", deviceList));
}

void destroy() {
  s_jsf.reset();
}

std::string execute(const std::string& Request) {
  return s_jsf->Execute(Request);
}

} // namespace jsf_mamgmt
