// Copyright (C) 2023, Xilinx Inc - All rights reserved
//
// Licensed under the Apache License, Version 2.0 (the "License"). You may
// not use this file except in compliance with the License. A copy of the
// License is located at
//
//      http://www.apache.org/licenses/LICENSE-2.0
//
// Unless required by applicable law or agreed to in writing, software
// distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
// WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
// License for the specific language governing permissions and limitations
// under the License.

#include "jsf_xrm.h"

#include <string_view>

#include <Logging.h>
#include <jsf_helpers.h>

using namespace nlohmann::json_literals;
using namespace std::literals;

namespace {

std::unique_ptr<jsf_int::Jsf> s_jsf;

class Device : public jsf_int::NodeLeaf {
private:
  unsigned m_deviceNum;

public:
  explicit Device(const pci_dev& PciDevice, unsigned DeviceNum) : NodeLeaf(jsf_helpers::DbdfString(PciDevice)), m_deviceNum(DeviceNum) {
  }

  virtual json Invoke(const json& /* Parameters */ = {}, bool /* Iteration */ = false) override final {
    json devices;
    json d;
    d["device_index"] = m_deviceNum;
    d["type"]         = "ma35";
    d["blocks"]       = R"([
        {
          "cuId         ": 0,
          "cuType       ": "IP Kernel",
          "kernelName   ": "scaler",
          "kernelAlias  ": "SCALER_AMA",
          "instanceName ": "scaler_1",
          "maxCapacity  ": 1990656000
        },
        {
          "cuId         ": 1,
          "cuType       ": "IP Kernel",
          "kernelName   ": "decoder",
          "kernelAlias  ": "DECODER_AMA",
          "instanceName ": "decoder_1",
          "maxCapacity  ": 1990656000
        },
        {
          "cuId         ": 2,
          "cuType       ": "IP Kernel",
          "kernelName   ": "encoder",
          "kernelAlias  ": "ENCODER_TYPE1_AMA",
          "instanceName ": "encoder_1",
          "maxCapacity  ": 501350400
        },
        {
          "cuId         ": 3,
          "cuType       ": "IP Kernel",
          "kernelName   ": "encoder",
          "kernelAlias  ": "ENCODER_TYPE2_AMA",
          "instanceName ": "encoder_2",
          "maxCapacity  ": 501350400
        },
        {
          "cuId         ": 4,
          "cuType       ": "IP Kernel",
          "kernelName   ": "encoder",
          "kernelAlias  ": "ENCODER_TYPE1_AMA",
          "instanceName ": "encoder_3",
          "maxCapacity  ": 501350400
        },
        {
          "cuId         ": 5,
          "cuType       ": "IP Kernel",
          "kernelName   ": "encoder",
          "kernelAlias  ": "ENCODER_TYPE2_AMA",
          "instanceName ": "encoder_4",
          "maxCapacity  ": 501350400
        },
        {
          "cuId         ": 6,
          "cuType       ": "IP Kernel",
          "kernelName   ": "lookahead",
          "kernelAlias  ": "LOOKAHEAD_AMA",
          "instanceName ": "lookahead_1",
          "maxCapacity  ": 995328000
        },
        {
          "cuId         ": 7,
          "cuType       ": "IP Kernel",
          "kernelName   ": "lookahead",
          "kernelAlias  ": "LOOKAHEAD_AMA",
          "instanceName ": "lookahead_2",
          "maxCapacity  ": 995328000
        },
        {
          "cuId         ": 8,
          "cuType       ": "IP Kernel",
          "kernelName   ": "ml",
          "kernelAlias  ": "ML_AMA",
          "instanceName ": "ml_1",
          "maxCapacity  ": 4000000
        }
      ])"_json;
    d["LogicGroups"]  = R"([
        {
          "cuId         ": [2, 6],
          "groupTypeName": "ENCODER_TYPE1"
        },
        {
          "cuId         ": [3, 6],
          "groupTypeName": "ENCODER_TYPE2"
        },
        {
          "cuId         ": [4, 7],
          "groupTypeName": "ENCODER_TYPE1"
        },
        {
          "cuId         ": [5, 7],
          "groupTypeName": "ENCODER_TYPE2"
        }
      ])"_json;
    return d;
  }
};

void DemoPopulate(jsf_int::Node* Node) {
  pci_dev dummy   = {};
  dummy.vendor_id = 0x10ee;
  dummy.device_id = 0x5070;
  dummy.bus       = 4;
  s_jsf->Insert(Node, s_jsf->Create<Device>(dummy, 0));
  dummy.bus = 5;
  s_jsf->Insert(Node, s_jsf->Create<Device>(dummy, 1));
}

} // namespace

namespace jsf_xrm {

void init(bool Demo) {
  s_jsf = jsf_int::CreateJsf();
  s_jsf->Insert(""sv, s_jsf->Create<jsf_int::NodeJson>("version", "0.0.1"));
  auto devicesNode = s_jsf->Insert(""sv, s_jsf->Create<jsf_int::NodeInternal>("devices"));
  if (Demo) {
    DemoPopulate(devicesNode);
  } else {
    for (const auto deviceId : {0x5070, 0x5071}) {
      auto devices = jsf_helpers::GetPciDevices(0x10ee, deviceId);
      for (const auto& device : devices) {
        int  device_num = -1;
        auto bdf        = jsf_helpers::DbdfString(device);

        for (unsigned dev_cnt = 0; dev_cnt < jsf_helpers::MAX_TRANS_DEVICES; ++dev_cnt) {
          auto          path = "/sys/class/misc/ama_transcoder"s + std::to_string(dev_cnt) + "/bus_id";
          std::ifstream stream(path);
          if (!stream.is_open()) {
            continue;
          }
          std::string transBdf;
          std::getline(stream, transBdf);
          if (bdf != transBdf) {
            continue;
          }
          device_num = dev_cnt;
          break;
        }
        if (device_num == -1) {
          logNotice("Transcoder device does not exist!");
          continue;
        }
        s_jsf->Insert(devicesNode, s_jsf->Create<Device>(device, device_num));
      }
    }
  }
  s_jsf->Insert(""sv, s_jsf->Create<jsf_helpers::NodeChildCount>("deviceCount", static_cast<jsf_int::NodeInternal*>(devicesNode)));
}

std::string execute(const std::string& Request) {
  return s_jsf->Execute(Request);
}

void destroy() {
}

} // namespace jsf_xrm
